!     THE SUBROUTINE HFUNIV CAN BE USED IN CALCULATIONS OF THE 
!     CHANDRASEKHAR H-FUNCTION FOR ISOTROPIC SCATTERING IN THE FOLLOWING
!     RANGES OF ARGUMENTS:   0 <= X <= 1   AND   0 <= OMEGA <= 1.
!     THIS ALGORITHM IMPLEMENTS THE DOUBLE EXPONENTIAL RULE. FOR SMALL
!     VALUES OF ARGUMENTS, THE ANALYTICAL APPROXIMATION IS USED.
!     THE CHANDRASEKHAR FUNCTION IS CALCULATED WITH ACCURACY OF
!     15 DECIMAL DIGITS OR BETTER.
!
!     INPUT PARAMETERS
!     X     - DIRECTIONAL VARIABLE 
!     OMEGA - ALBEDO 
!
!     OUTPUT PARAMETERS
!     H     - VALUE OF THE CHANDRASEKHAR FUNCTION
!     INDEX - FLAG
!             INDEX=0 FAILURE. ARGUMENTS ARE OUT OF THE RANGE OF
!                     VALIDITY OR THE DE-RULE IS NOT CONVEGENT
!             INDEX=1 SUCCESS
!
!     A. Jablonski
!     Institute of Physical Chemistry, Polish Academy of Sciences
!     Kasprzaka 44/52, 01-224 Warsaw, Poland
!     e-mail: ajablonski@ichf.edu pl
!     23 January 2020
!
!     ##################################################################
      SUBROUTINE HFUNIV(X,OMEGA,H,INDEX)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      PARAMETER (PI=3.14159265358979323846264338328D0)
      H=0.0D0
      INDEX=0
      IF (X .LT. 0.0D0 .OR. X .GT. 1.0D0) RETURN
      IF (OMEGA .LT. 0.0D0 .OR. OMEGA .GT. 1.0D0) RETURN
      INDEX=1
      IF (X .EQ. 0.0D0 .OR. OMEGA .EQ. 0.0D0) THEN
        H=1.0D0
        RETURN
      ENDIF
      OMEGA1=DEXP(-(0.493D0*DLOG(X)+17.5D0))
      IF (OMEGA .LE. OMEGA1) THEN         
        R=(PI/2.0D0)*DLOG((1.0D0+X)/X)
        HLOG=-(X*OMEGA/PI)*R
        H=DEXP(-HLOG)      
        RETURN
      ENDIF
      CALL HFDE(X,OMEGA,H,INDEX)
      RETURN
      END      
      SUBROUTINE HFDE(X,OMEGA,H,INDEX)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      COMMON XMU0,OMEGA0
      EXTERNAL FUNDE
      XMU0=X
      OMEGA0=OMEGA
      TMAX=6.0D0
      EPS=1.0D-15
      CALL TRAPEZ(FUNDE,-TMAX,TMAX,EPS,H,R,INDEX)
      RETURN
      END
      FUNCTION FUNDE(T)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      COMMON XMU,OMEGA
      DIMENSION CC(12)
      PARAMETER (PI=3.14159265358979323846264338328D0,PI2=PI/2.0D0)
      DATA CC/ 0.333333333333333333333D+00,-0.200000000000000000000D+00,&
     &         0.142857142857142857143D+00,-0.111111111111111111111D+00,&
     &         0.909090909090909090909D-01,-0.769230769230769230769D-01,&
     &         0.666666666666666666667D-01,-0.588235294117647058824D-01,&
     &         0.526315789473684210526D-01,-0.476190476190476190476D-01,&
     &         0.434782608695652173913D-01,-0.400000000000000000000D-01/      
      X=DEXP(PI2*DSINH(T))
      A=X*PI2*DCOSH(T)
      IF (X .LT. 0.15D0) THEN
        W=X*X 
        ARGUM=CC(11)*W
        DO 30 J=1,10
        ARGUM=(ARGUM+CC(11-J))*W
  30    CONTINUE          
      ELSE
        ARGUM=1.0D0-DATAN(X)/X
      ENDIF
      IF (OMEGA .GT. 0.999D0) THEN
        ARGUM=DNINT((1.0D0-OMEGA)*1.0D15)*1.0D-15+OMEGA*ARGUM 
      ELSE
        ARGUM=1.0D0-OMEGA+OMEGA*ARGUM
      ENDIF      
      B=DLOG(ARGUM)
      C=1.0D0/(1.0D0+XMU*XMU*X*X)      
      FUNDE=C*B*A      
      RETURN
      END
      SUBROUTINE TRAPEZ(FF,A,B,E0,HF,R,I1)
      IMPLICIT DOUBLE PRECISION(A-H,O-Z)
      COMMON XMU,OMEGA 
      PARAMETER (PI=3.14159265358979323846264338328D0)       
	  I1=1
      K=1
      H=(B-A)
      S=(FF(A)+FF(B))/2.0D0
	  V=H*S
	  HFPREV=DEXP(-(XMU/PI)*V)  	  
 100  CONTINUE
      IF (K .GT. 15) THEN
	    I1=0
		RETURN
	  ENDIF
      K=K+1
	  H=H/2.0D0
	  J=2**(K-2)
      DO 10 I=1,J
	  X=A+(2*I-1)*H
	  S=S+FF(X)
  10  CONTINUE
	  V=H*S
      HF=DEXP(-(XMU/PI)*V)
	  R=DABS((HF-HFPREV)/HF)
	  IF (R .GT. E0) THEN
	    HFPREV=HF
		GOTO 100
	  ENDIF
      RETURN
      END
